// DCC Decoder for DS-DCC decode
// By yaasan
// Based on Nicolas's sketch http://blog.nicolas.cx
// Inspired by Geoff Bunza and his 17 Function DCC Decoder & updated library
//
//
// Debug serial output available on the serial port at baud 115200, aka Tools -> Serial Monitor
//

//なごでんさんのｍｐ3スケッチ　ＭＰＶ5＿ＤＬ110を参考に、走行信号、ファンクション受信部を書き直し

//fujigayaさんのDesktop Station for Smart Controller 181027を参考に、サウンドボックスの通信部分作成

//KATO SOSUNDBOX 制御

//CVack機能追加

//#include <NmraDcc.h>
#include "NmraDcc_x.h"
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below

//181027 DesktopStation for Kato SmartController fujigaya2
#include "sc_function.h"
#include "Functions.h"

//Task Schedule
unsigned long gPreviousL1 = 0; //  300ms interval

//各種設定、宣言


#define DECODER_ADDRESS 3
#define DCC_ACK_PIN 9	//if defined enables the ACK pin functionality. Comment out to disable.

#define MOTOR_LIM_MAX 255

#define CV_VSTART		2
#define CV_ACCRATIO		3
#define CV_DECCRATIO	4

#define CV_FUNCSTATE 200


//使用クラスの宣言
NmraDcc	 Dcc;
DCC_MSG	 Packet;

//Task Schedule
unsigned long gPreviousL5 = 0;

//モータ制御関連の変数
uint32_t gPwmLPF_buf = 0;
uint32_t gPwmLPF2_buf = 0;
uint8_t gPwmDirv = 128;
uint8_t gCV29_Vstart = 0;
uint8_t gCV29Direction = 0;
uint16_t gSpeedCmd = 0;
uint16_t gPrevSpeed = 0;

uint8_t gCV1_SAddr = 3;
uint8_t gCVx_LAddr = 3;
uint8_t gCV2_Vstart = 32;
uint8_t gCV3_AccRatio = 32;
uint8_t gCV4_DecRatio = 32;

//Internal variables and other.
#if defined(DCC_ACK_PIN)
const int DccAckPin = DCC_ACK_PIN ;
#endif

/* Function assigned pins */
const int FunctionPin0 = 3;
const int FunctionPin1 = 4;
const int FunctionPin2 = 5;
const int FunctionPin3 = 6;
const int FunctionPin4 = 7;
const int FunctionPin5 = 8;

struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};

CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_29_CONFIG, 128 },	 //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
  {CV_VSTART, 16},
  {CV_ACCRATIO, 0},
  {CV_DECCRATIO, 0},
};

//コマンド変更確認用変数
char function_s[] =  {0,0,0,0,0,0,0,0,0};
uint8_t spd_s = 0;
uint8_t dir_s = 10;

void(* resetFunc) (void) = 0;  //declare reset function at address 0

uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);


void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

	resetCVToDefault();	 
	Serial.println("Resetting...");
	delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying

	resetFunc();
};


void resetCVToDefault()
{
	//CVをデフォルトにリセット
	Serial.println("CVs being reset to factory defaults");
	
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
};



extern void	   notifyCVChange( uint16_t CV, uint8_t Value){
   //CVが変更されたときのメッセージ
   Serial.print("CV "); 
   Serial.print(CV); 
   Serial.print(" Changed to "); 
   Serial.println(Value, DEC);
};

void notifyCVAck(void)
{
  Serial.println("notifyCVAck");
  analogWrite( DccAckPin, 255 );
  delay( 6 ); 
  analogWrite( DccAckPin, 0 );
}


void setup()
{

  // put your setup code here, to run once:
  //スマートコントローラ側通信
  sc_init_serial();
  
	uint8_t cv_value;
	
	//シリアル通信開始
	Serial.begin(115200);
  Serial.setTimeout(500);

  Serial.println("--------------------------------------");
  Serial.println("Smile Decorder for Kato SoundBox  ");
  Serial.println("--------------------------------------");
  Serial.println("100 Ready");

  //Reset task
  //gPreviousL1 = millis();

	//DCCの応答用負荷ピン
	
	#if defined(DCCACKPIN)
	Setup ACK Pin
	pinMode(DccAckPin,OUTPUT);
	digitalWrite(DccAckPin, 0);
	#endif
  


  #if !defined(DECODER_DONT_DEFAULT_CV_ON_POWERUP)
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF ){	 //if eeprom has 0xFF then assume it needs to be programmed
	  Serial.println("CV Defaulting due to blank eeprom");
	  notifyCVResetFactoryDefault();
	  
   } else{
	 Serial.println("CV Not Defaulting");
   }
  #else
	 Serial.println("CV Defaulting Always On Powerup");
	 notifyCVResetFactoryDefault();
  #endif 
  
   
   
  // Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
  Dcc.pin(0, 2, 0);
  
  // Call the main DCC Init function to enable the DCC Receiver
  Dcc.init( MAN_ID_DIY, 100,   FLAGS_MY_ADDRESS_ONLY , 0 ); 

  //Reset task
  gPreviousL5 = millis();
  
  //Init CVs
  gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
  gCV2_Vstart = Dcc.getCV( CV_VSTART ) ;
  gCV3_AccRatio = Dcc.getCV( CV_ACCRATIO ) ;
  gCV4_DecRatio = Dcc.getCV( CV_DECCRATIO ) ;
  gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8) + Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
  
  
  Serial.print("CV1(ShortAddr): ");
  Serial.println(gCV1_SAddr);
  Serial.print("CV17/18(LongAddr): ");
  Serial.println(gCVx_LAddr);
  Serial.print("CV2(Vstart): ");
  Serial.println(gCV2_Vstart);
  Serial.print("CV3(AccRatio): ");
  Serial.println(gCV3_AccRatio);
  Serial.print("CV4(DecRatio): ");
  Serial.println(gCV4_DecRatio);  
  Serial.println("Ready");
  
}
  //アドレスをＤＣＣに変換
  word loco_address = 49152 + gCV1_SAddr ;

void loop(){
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	

	if( (millis() - gPreviousL5) >= 100)
	{
		//Motor drive control
		Motor_Control();
		
		//Reset task
		gPreviousL5 = millis();
		
	}


}

/* Motor control Task (10Hz) */
void Motor_Control()
{

  Serial.print("Speed - ADR: ");
  Serial.print(loco_address);
  Serial.print(", SPD: ");
  Serial.print(gSpeedCmd * 4);
  Serial.print(", DIR: ");
  Serial.print(gPwmDirv);
  Serial.println("");

  if(gPwmDirv != dir_s)
  {
    if(gPwmDirv == 0)//0 or 128
    {
      //0 forward
      sc_SetLocoDirection(loco_address, 1) ;
    }
    else
    {
      //128 reverse
      sc_SetLocoDirection(loco_address, 2) ;
    }
    dir_s = gPwmDirv; 
  }
  else if(gSpeedCmd != spd_s)
  {
    sc_SetLocoSpeedEx(loco_address, gSpeedCmd * 4, 0);
    spd_s = gSpeedCmd;
  }
/*
  sc_SetLocoDirection(loco_address, inDirection) ;
  sc_SetLocoSpeedEx(loco_address, inSpeed, 0);
*/
}

//DCC速度信号の受信によるイベント
extern void notifyDccSpeed( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Speed, DCC_DIRECTION Dir, DCC_SPEED_STEPS SpeedSteps )
{
	
	
  
  uint16_t aSpeedRef = 0;
  
  //速度値の正規化(255を100%とする処理)
  if( Speed >= 1)
  {
    aSpeedRef = ((Speed - 1) * 255) / SpeedSteps;
  }
  else
  {
    //緊急停止信号受信時の処理 //Nagoden comment 2016/06/11
#ifdef DEBUG
    Serial.println("***** Emagency STOP **** ");
#endif
    aSpeedRef = 0;
  }

  
  gSpeedCmd = aSpeedRef;
  gPwmDirv = Dir;
  if ( gCV29Direction > 0)
  {
     if ( Dir == DCC_DIR_FWD)
     {
          gPwmDirv = 128;
     }
    else
    {
          gPwmDirv = 0;     
    }     
  }
  else
  {
     if ( Dir == DCC_DIR_FWD)
     {
          gPwmDirv = 0;
     }
    else
    {
         gPwmDirv =128;     
    }     
  }

#ifdef DEBUG
	// デバッグメッセージ
	Serial.print("Speed - ADR: ");
	Serial.print(Addr);
	Serial.print(", SPD: ");
	Serial.print(Speed);
	Serial.print(", DIR: ");
	Serial.print(gPwmDirv);
	Serial.print(", MAX: ");
	Serial.println(MaxSpeed);
#endif
}

//ファンクション信号受信のイベント
extern void notifyDccFunc( uint16_t Addr, DCC_ADDR_TYPE AddrType, FN_GROUP FuncGrp, uint8_t FuncState)
{
	switch(FuncGrp)
	{
		case FN_0_4:	//Function Group 1 F0 F4 F3 F2 F1   
			exec_function( 0, (FuncState & FN_BIT_00)>>4 );
			exec_function( 1, (FuncState & FN_BIT_01));
			exec_function( 2, (FuncState & FN_BIT_02)>>1);
			exec_function( 3, (FuncState & FN_BIT_03)>>2 );
			exec_function( 4, (FuncState & FN_BIT_04)>>3 );
			break;

		case FN_5_8:	//Function Group 1 S FFFF == 1 F8 F7 F6 F5	&  == 0	 F12 F11 F10 F9 F8
			exec_function( 5, (FuncState & FN_BIT_05));
			exec_function( 6, (FuncState & FN_BIT_06)>>1 );
			exec_function( 7, (FuncState & FN_BIT_07)>>2 );
			exec_function( 8, (FuncState & FN_BIT_08)>>3 );
			break;

		case FN_9_12:
			//exec_function( 9, (FuncState & FN_BIT_09));
			//exec_function( 10, (FuncState & FN_BIT_10)>>1 );
			//exec_function( 11, (FuncState & FN_BIT_11)>>2 );
			//exec_function( 12, (FuncState & FN_BIT_12)>>3 );
			break;

		case FN_13_20:	 //Function Group 2 FuncState == F20-F13 Function Control
			//exec_function( 13, (FuncState & FN_BIT_13));
			//exec_function( 14, (FuncState & FN_BIT_14)>>1 );
			//exec_function( 15, (FuncState & FN_BIT_15)>>2 );
			//exec_function( 16, (FuncState & FN_BIT_16)>>3 );
			break;

		case FN_21_28:
			break;
			
	}
}


void exec_function (int function, int FuncState)
{
  if(function_s[function] != FuncState){
    
    Serial.print(" DCC_adress: ");
    Serial.print(loco_address);
    Serial.print(" Function: ");
    Serial.print(function);
    Serial.print(" : ");
    Serial.println(FuncState);
    function_s[function] = FuncState;
    //ファンクション（音を出す）
    //０の時だけ、この関数内で処理
    if(function == 0)
    {
      //トグル計算
      gCV29Direction = gCV29Direction * (-1) + 1;
    }
    else
    {
      sc_SetLocoFunction(loco_address, function, FuncState);
    }
  }
}
